"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _errors = require("./errors");
// #region Types

/** Infer the type of a dictionary */

/** Infer the type of additional schema dictionaries */

/** Type which will produce a flattened document of only scalars with dictionaries */

/** Query Filter */

/** Sort criteria */

// #endregion

/** A query object */
class Query {
  /** Connection instance to run query on */

  /** Schema used for query */

  /** File to run query against */

  /** Log handler instance used for diagnostic logging */

  /** String to use as selection criteria in query */

  /** String to use as sort criteria in query */

  /** Sort criteria passed to constructor */

  /** Limit the result set to this number of items */

  /** Skip this number of items in the result set */

  /** Specify the projection attribute in result set */

  /** Number of conditions in the query */
  conditionCount = 0;
  constructor(connection, schema, file, logHandler, selectionCriteria, options = {}) {
    const {
      sort,
      limit,
      skip,
      projection
    } = options;
    this.connection = connection;
    this.schema = schema;
    this.file = file;
    this.logHandler = logHandler;
    this.limit = limit;
    this.skip = skip;
    this.projection = projection ?? null;
    this.selection = this.formatSelectionCriteria(selectionCriteria);
    this.sortCriteria = sort;
    this.sort = this.formatSortCriteria(sort);
  }

  /** Execute query */
  async exec(options = {}) {
    const {
      maxReturnPayloadSize,
      requestId,
      userDefined
    } = options;
    let queryCommand = `select ${this.file}`;
    if (this.selection != null) {
      queryCommand = `${queryCommand} with ${this.selection}`;
    }
    if (this.sort != null) {
      queryCommand = `${queryCommand} ${this.sort}`;
    }
    await this.validateQuery(queryCommand, requestId);
    const projection = this.projection != null && this.schema != null ? this.schema.transformPathsToDbPositions(this.projection) : null;
    const executionOptions = {
      filename: this.file,
      queryCommand,
      ...(this.skip != null && {
        skip: this.skip
      }),
      ...(this.limit != null && {
        limit: this.limit
      }),
      projection
    };
    const setupOptions = {
      ...(maxReturnPayloadSize != null && {
        maxReturnPayloadSize
      }),
      ...(requestId != null && {
        requestId
      }),
      ...(userDefined && {
        userDefined
      })
    };
    this.logHandler.debug(`executing query "${queryCommand}"`);
    const data = await this.connection.executeDbSubroutine('find', executionOptions, setupOptions);
    return {
      count: data.count,
      documents: data.documents
    };
  }

  /**
   * Format the selection criteria object into a string to use in multivalue query
   * @throws {@link TypeError} The value of the $or property must be an array
   * @throws {@link TypeError} Invalid conditional operator specified
   */
  formatSelectionCriteria(criteria) {
    const criteriaProperties = Object.keys(criteria);
    if (criteriaProperties.length === 0) {
      return null;
    }
    const andConditions = Object.entries(criteria).map(([queryProperty, queryValue]) => {
      if (queryProperty === '$or' || queryProperty === '$and') {
        if (!Array.isArray(queryValue) || queryValue.length === 0) {
          throw new TypeError(`The value of the ${queryProperty} property must be an array`);
        }
        const conditions = queryValue.map(queryObj => this.formatSelectionCriteria(queryObj));
        if (conditions.length === 1) {
          return conditions[0];
        }
        const joinWord = queryProperty === '$or' ? ' or ' : ' and ';
        return `(${conditions.join(joinWord)})`;
      }
      if (Array.isArray(queryValue)) {
        // assume $in operator if queryValue is an array
        return this.formatConditionList(queryProperty, '=', queryValue, 'or');
      }
      if (typeof queryValue !== 'object' || queryValue === null) {
        // assume equality if queryValue is not an object or is equal null
        return this.formatCondition(queryProperty, '=', queryValue);
      }

      // if query value is an object then it should contain one or more pairs of operator and value
      const operatorConditions = Object.entries(queryValue).map(([operator, mvValue]) => {
        switch (operator) {
          case '$eq':
            return this.formatCondition(queryProperty, '=', mvValue);
          case '$gt':
            return this.formatNonNullableCondition(queryProperty, '>', mvValue);
          case '$gte':
            return this.formatNonNullableCondition(queryProperty, '>=', mvValue);
          case '$lt':
            return this.formatNonNullableCondition(queryProperty, '<', mvValue);
          case '$lte':
            return this.formatNonNullableCondition(queryProperty, '<=', mvValue);
          case '$ne':
            return this.formatCondition(queryProperty, '#', mvValue);
          case '$contains':
            this.validateLikeCondition(mvValue);
            return this.formatNonNullableCondition(queryProperty, 'like', `...'${mvValue}'...`);
          case '$startsWith':
            this.validateLikeCondition(mvValue);
            return this.formatNonNullableCondition(queryProperty, 'like', `'${mvValue}'...`);
          case '$endsWith':
            this.validateLikeCondition(mvValue);
            return this.formatNonNullableCondition(queryProperty, 'like', `...'${mvValue}'`);
          case '$in':
            return this.formatConditionList(queryProperty, '=', mvValue, 'or');
          case '$nin':
            return this.formatConditionList(queryProperty, '#', mvValue, 'and');
          default:
            // unknown operator
            throw new TypeError('Invalid conditional operator specified');
        }
      });
      return operatorConditions.length === 1 ? operatorConditions[0] : `(${operatorConditions.join(' and ')})`;
    });
    return andConditions.length === 1 ? andConditions[0] : `(${andConditions.join(' and ')})`;
  }

  /** Format the sort criteria object into a string to use in multivalue query */
  formatSortCriteria(criteria) {
    if (criteria == null || criteria.length === 0) {
      return null;
    }
    return criteria.map(sortEntry => {
      const [sortProperty, sortOrder] = sortEntry;
      const dictionaryId = this.getDictionaryId(sortProperty);
      let byClause;
      if (sortOrder === -1) {
        byClause = 'by.dsnd';
      } else {
        byClause = 'by';
      }
      return `${byClause} ${dictionaryId}`;
    }).join(' ');
  }

  /**
   * Format a conditional expression which prohibits null values
   * @throws {@link InvalidParameterError} An invalid parameter was passed to the function
   */
  formatNonNullableCondition(property, operator, value) {
    if (value == null) {
      throw new _errors.InvalidParameterError({
        parameterName: 'value'
      });
    }
    return this.formatCondition(property, operator, value);
  }

  /** Format a conditional expression */
  formatCondition(property, operator, value) {
    this.conditionCount += 1;
    const dictionaryId = this.getDictionaryId(property);
    return `${dictionaryId} ${operator} ${this.formatConstant(property, value)}`;
  }

  /**
   * Format a list of conditional expressions
   * @throws {@link InvalidParameterError} An invalid parameter was passed to the function
   */
  formatConditionList(property, operator, valueList, joinString) {
    if (valueList.length === 0) {
      throw new _errors.InvalidParameterError({
        parameterName: 'valueList'
      });
    }
    const conditionList = valueList.map(value => this.formatCondition(property, operator, value));
    return conditionList.length === 1 ? conditionList[0] : `(${conditionList.join(` ${joinString} `)})`;
  }

  /**
   * Format a constant for use in queries
   * @throws {@link Error} Passed constant parameter contains both single and double quotes
   */
  formatConstant(property, constant) {
    if (constant == null) {
      // convert null to empty string
      return '""';
    }
    const constantToFormat = this.transformToQuery(property, constant);
    if (typeof constantToFormat === 'string' && constantToFormat.includes(`'`) && constantToFormat.includes(`"`)) {
      // cannot query if string has both single and double quotes in it
      throw new Error('Query constants cannot contain both single and double quotes');
    }
    const quoteCharacter = typeof constantToFormat === 'string' && constantToFormat.includes(`"`) ? `'` : `"`;
    return `${quoteCharacter}${constantToFormat}${quoteCharacter}`;
  }

  /**
   * Get a dictionary id at a given schema path
   * @throws {link InvalidParameterError} Nonexistent schema property or property does not have a dictionary specified
   */
  getDictionaryId(property) {
    const dictionaryTypeDetail = this.schema?.dictPaths.get(property);
    if (dictionaryTypeDetail == null) {
      throw new _errors.InvalidParameterError({
        message: 'Nonexistent schema property or property does not have a dictionary specified',
        parameterName: 'property'
      });
    }
    return dictionaryTypeDetail.dictionary;
  }

  /** Transform query constant to internal u2 format (if applicable) */
  transformToQuery(property, constant) {
    const dictionaryTypeDetail = this.schema?.dictPaths.get(property);
    if (dictionaryTypeDetail == null) {
      throw new _errors.InvalidParameterError({
        message: 'Nonexistent schema property or property does not have a dictionary specified',
        parameterName: 'property'
      });
    }
    return dictionaryTypeDetail.dataTransformer.transformToQuery(constant);
  }

  /** Validate the query before execution */
  async validateQuery(query, requestId) {
    const {
      maxSort,
      maxWith,
      maxSentenceLength
    } = await this.connection.getDbLimits({
      requestId
    });

    /*
        For some reason, UDTEXECUTE (which is used by this library) appears to have a sentence length limit which is one character
        less than the sentence length limit used by EXECUTE.  The sentence length returned in the LIMIT values represent that of
        EXECUTE and not UDTEXECUTE.  The below code reduces the length by 1 to accommodate for that discrepancy.
      */
    if (query.length > maxSentenceLength - 1) {
      throw new _errors.QueryLimitError({
        message: 'Query length exceeds maximum sentence length of database server'
      });
    }
    if (this.sortCriteria != null && this.sortCriteria.length > maxSort) {
      throw new _errors.QueryLimitError({
        message: 'Query sort criteria exceeds maximum number of sort criteria of database server'
      });
    }
    if (this.conditionCount > maxWith) {
      throw new _errors.QueryLimitError({
        message: 'Query condition count exceeds maximum number of query conditions of database server'
      });
    }
  }

  /** Validate that a "like" condition does not contain quotes and is not null */
  validateLikeCondition(value) {
    if (value == null) {
      throw new _errors.InvalidParameterError({
        parameterName: 'value'
      });
    }
    const stringValue = String(value);
    if (stringValue.includes(`'`) || stringValue.includes(`"`)) {
      // cannot query if like condition has single or double quotes in it
      throw new Error('$contains, $startsWith, and $endsWith queries cannot contain single or double quotes in the conditional value');
    }
  }
}
var _default = exports.default = Query;